import _Object$keys from 'babel-runtime/core-js/object/keys';
import _typeof from 'babel-runtime/helpers/typeof';
import _classCallCheck from 'babel-runtime/helpers/classCallCheck';
import _Symbol from 'babel-runtime/core-js/symbol';
import invariant from 'invariant';
import { Subject } from 'rxjs';
import toMap from './util/toMap';
import createReducerForStore from './util/createReducerForStore';
import { Collection } from './Collection';

import { Iterable, OrderedMap, Set } from 'immutable';

import { STORE_INSERT, STORE_REMOVE, STORE_FILTER, STORE_UPDATE } from './constants/StoreConstants';

var missing = _Symbol('missingSubject');

export var Store = function () {
  Store.isStore = function isStore(obj) {
    return (typeof obj === 'undefined' ? 'undefined' : _typeof(obj)) === 'object' && obj instanceof Store;
  };

  function Store(identifier) {
    var opts = arguments.length <= 1 || arguments[1] === undefined ? {} : arguments[1];

    _classCallCheck(this, Store);

    invariant(typeof identifier === 'string', 'Store: `identifier` is expected to be a string.');

    this[missing] = {
      subject: new Subject(),
      cache: {}
    };

    this.identifier = identifier;
    this.opts = opts;
    this.dependencies = {};
    this.hooks = {};

    this.insert = this.insert.bind(this);
    this.remove = this.remove.bind(this);
    this.filter = this.filter.bind(this);
    this.update = this.update.bind(this);
  }

  Store.prototype.pre = function pre(transformer) {
    invariant(typeof transformer === 'function', 'Store: `transformer` is expected to be a function.');

    this.hooks.pre = transformer;
    return this;
  };

  Store.prototype.dependsOn = function dependsOn(identifier) {
    var dependencies = this.dependencies;


    invariant(typeof identifier === 'string', 'Store: `identifier` is expected to be a string.');
    invariant(!dependencies[identifier], 'Store: There is already an existing dependency to the store `' + identifier + '`.');

    for (var _len = arguments.length, args = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
      args[_key - 1] = arguments[_key];
    }

    invariant(args.length <= 2 && args.length >= 1, 'Store: `dependsOn` is expected to receive `setter` or `getter` and `setter`.');

    var dependency = void 0;
    if (args.length === 1) {
      var setter = args[0];

      invariant(typeof setter === 'function', 'Store: `setter` is expected to be a function.');

      dependency = { identifier: identifier, setter: setter };
    } else {
      var getter = args[0];
      var _setter = args[1];

      invariant(typeof getter === 'function', 'Store: `getter` is expected to be a function.');
      invariant(typeof _setter === 'function', 'Store: `setter` is expected to be a function.');

      dependency = { identifier: identifier, getter: getter, setter: _setter };
    }

    dependencies[identifier] = dependency;
    return this;
  };

  Store.prototype.post = function post(transformer) {
    invariant(typeof transformer === 'function', 'Store: `transformer` is expected to be a function.');

    this.hooks.post = transformer;
    return this;
  };

  Store.prototype.observeMissing = function observeMissing() {
    return this[missing].subject.asObservable();
  };

  Store.prototype.useCollection = function useCollection(col) {
    invariant(col && col.prototype instanceof Collection, 'Store: `col` is expected to be a Collection.');

    this.collection = col;
    return this;
  };

  Store.prototype.getIdentifier = function getIdentifier() {
    return this.identifier;
  };

  Store.prototype.getPre = function getPre() {
    return this.hooks.pre || function (x) {
      return toMap(x);
    };
  };

  Store.prototype.getDependencies = function getDependencies() {
    return this.dependencies;
  };

  Store.prototype.getPost = function getPost() {
    return this.hooks.post;
  };

  Store.prototype.getReducer = function getReducer() {
    if (this.reducer && typeof this.reducer === 'function') {
      return this.reducer;
    }

    this.reducer = createReducerForStore(this);
    return this.reducer;
  };

  Store.prototype.createCollection = function createCollection() {
    var CollectionClass = this.collection || Collection;

    var res = new CollectionClass();
    res.dependencies = _Object$keys(this.dependencies);

    return res;
  };

  Store.prototype._missing = function _missing(ids) {
    var identifier = arguments.length <= 1 || arguments[1] === undefined ? null : arguments[1];
    var _missing2 = this[missing];
    var cache = _missing2.cache;
    var subject = _missing2.subject;


    invariant(Set.isSet(ids) && ids.every(function (x) {
      return typeof x === 'string';
    }), 'Store: `ids` is expected to be an Immutable.Set of ids (string).');

    cache[identifier] = ids;

    var missingIds = _Object$keys(cache).reduce(function (acc, key) {
      var _set = cache[key];
      return _set ? acc.union(_set) : acc;
    }, new Set());

    subject.next(missingIds);
  };

  Store.prototype.insert = function insert(payload, groupId) {
    invariant(payload && (Iterable.isKeyed(payload) || Iterable.isIterable(payload) && payload.every && payload.every(Iterable.isKeyed)), 'Store: `payload` is expected to be a keyed iterable or an iterable containing keyed iterables.');
    var identifier = this.identifier;


    return {
      type: STORE_INSERT,
      identifier: identifier,
      payload: payload,
      groupId: groupId
    };
  };

  Store.prototype.remove = function remove(payload) {
    invariant(payload && (typeof payload === 'string' || Iterable.isKeyed(payload) && typeof payload.get('id') === 'string'), 'Store: `payload` is expected to be an id or a keyed iterable containing an id.');
    var identifier = this.identifier;


    return {
      type: STORE_REMOVE,
      identifier: identifier,
      payload: payload
    };
  };

  Store.prototype.filter = function filter(selector) {
    invariant(typeof selector === 'function', 'Store: `selector` is expected to be a function.');
    var identifier = this.identifier;


    return {
      type: STORE_FILTER,
      identifier: identifier,
      selector: selector
    };
  };

  Store.prototype.update = function update(selector) {
    invariant(typeof selector === 'function', 'Store: `selector` is expected to be a function.');
    var identifier = this.identifier;


    return {
      type: STORE_UPDATE,
      identifier: identifier,
      selector: selector
    };
  };

  return Store;
}();

export default function createStore(identifier, opts) {
  return new Store(identifier, opts);
}